/*****************************************************************************/

/*
 *	keypad.c -- simple keypad driver.
 *
 *	(C) Copyright 2001, Greg Ungerer (gerg@snapgear.com)
 *	(C) Copyright 2001, Lineo Inc. (www.lineo.com) 
 * 
 * Modified 2005 NZG EMAC.Inc to load in the 2.6 kernel and support the
 * SoM-100ES. Also modified to support loading the keypad matrix via IOCTL
 * Low level calls are broken into inlines in keypad.h
 */

/*****************************************************************************/

#include <linux/module.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/keypad.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/param.h>
#include <asm/coldfire.h>
#include <asm/mcfsim.h>

MODULE_DESCRIPTION( "generic keypad driver" );
MODULE_AUTHOR( "Greg Ungerer <gerg@snapgear.com>,NZG <ngustavson@emacinc.com>" );
MODULE_LICENSE("Dual BSD/GPL");

#define DRIVER_NAME "keypad"


/*****************************************************************************/

/**
 * debug statements
 */
//#define KEYPAD_DEBUG 1

/*
 *	Place to store key codes at interrupt/poll time.
 */
#define	KEYPAD_BUFSIZE		8

unsigned char	keypad_buf[KEYPAD_BUFSIZE];
int						keypad_bufhead;
int						keypad_bufcount;
int						keypad_isopen;
int						keypad_blocking;
static int			new_input;
int 						countdown;

static struct cdev *keypad_cdev;
DECLARE_WAIT_QUEUE_HEAD(keypad_waitq_head);

static struct timer_list poll_timer;

//simper and quicker to copy a single dimension from user space
char keypad_matrix[KEYPAD_ROWS*KEYPAD_COLUMNS];

/*****************************************************************************/

int keypad_open(struct inode *inode, struct file *filp)
{
#if KEYPAD_DEBUG
	printk("keypad_open()\n");
#endif

	keypad_blocking = (filp->f_flags & O_NONBLOCK) ? 0 : 1;
	keypad_isopen++;
	return(0);
}

/*****************************************************************************/

int keypad_release(struct inode *inode, struct file *filp)
{
#if KEYPAD_DEBUG
	printk("keypad_close()\n");
#endif
	keypad_isopen = 0;
	
	return 0;
}

/*****************************************************************************/
static ssize_t keypad_read (struct file *file, char __user *buf, size_t count, loff_t *loff)
{
#ifndef POLLING
	unsigned long	flags;
	int		i;
#endif
	if (count == 0)
		return 0;

#if KEYPAD_DEBUG
	printk("keypad_read(buf=%x,count=%d)\n", (int) buf, count);
#endif


#if POLLING
	del_timer(&poll_timer);

	put_user(keypad_buf[keypad_bufhead], buf);
	
	if (++keypad_bufhead >= KEYPAD_BUFSIZE)
		keypad_bufhead = 0;
	keypad_bufcount--;
	if (keypad_bufcount < 0)
		keypad_bufcount = 0;

	poll_timer.expires = jiffies + TIMEOUT;
	add_timer(&poll_timer);
	return(1);
#else

if(keypad_blocking)
		wait_event_interruptible(keypad_waitq_head,(keypad_bufcount));	
		//wait_event_interruptible_timeout(keypad_waitq_head,(keypad_bufcount),HZ);	


	save_flags(flags); cli();	
	if (count > keypad_bufcount)
		count = keypad_bufcount;
	for (i = 0; i < count; i++) {
		put_user(keypad_buf[keypad_bufhead], buf);
		if (++keypad_bufhead >= KEYPAD_BUFSIZE)
			keypad_bufhead = 0;
	}
	keypad_bufcount -= i;
	restore_flags(flags);
	return(i);
#endif
}

/*****************************************************************************/
static unsigned int keypad_poll (struct file *file, struct poll_table_struct *wait)
{
	int mask=0;
	poll_wait(file, &keypad_waitq_head, wait);					
	if (keypad_bufcount > 0)
		mask = POLLIN|POLLRDNORM;
	return mask;
}




/*****************************************************************************
 * inlines used in keypad_timer
 */
inline void restart_timer(void){
	poll_timer.expires = jiffies + TIMEOUT;
	add_timer(&poll_timer);
}

inline void timer_exit(void){
	#ifdef POLLING
	restart_timer();
	#else
	KEYPAD_UNMASK_IRQ();
	#endif
}

inline void enqueue_key(u8 key){
	int pos;
	pos = (keypad_bufhead + keypad_bufcount);
	while (pos >= KEYPAD_BUFSIZE)
	pos -= KEYPAD_BUFSIZE;
	keypad_buf[pos] = key;
	keypad_bufcount++;
	if (keypad_bufcount >= KEYPAD_BUFSIZE)
		keypad_bufcount--;
}


/*****************************************************************************/
static void keypad_timer(void) {

	del_timer(&poll_timer);

if(countdown){
	if(KEYHELD){
			countdown--;
			restart_timer();
		}
	else{
		countdown=0;
		timer_exit();
		return;
	}	
}
else{//countdown is zero
	unsigned short pos = GETKEYPOS();	
	char key = keypad_matrix[ROW(pos)*KEYPAD_COLUMNS+COLUMN(pos)];

#if KEYPAD_DEBUG
	printk("keypad_read(row = %u  column = %u-> key = %c)\n", ROW(pos),COLUMN(pos),key);
#endif

 	enqueue_key(key);
	new_input = 1;
	wake_up_interruptible(&keypad_waitq_head);
	if(KEYHELD){
			countdown=REPEAT_DELAY;
			restart_timer();
		}
	else{
		timer_exit();
		return;
	}		
}//countdown is zero

}

/*****************************************************************************/

int keypad_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int	rc = 0;

	switch (cmd) {
	case SETKEYARRAY:					
		if(access_ok(VERIFY_READ,(char *)arg,sizeof(keypad_matrix)))
    		copy_from_user(keypad_matrix,(char *)arg,sizeof(keypad_matrix));
		else
   			 return -EINVAL;
		break;
	case GETKEYARRAY:
	if(access_ok(VERIFY_WRITE,(char *)arg,sizeof(keypad_matrix)))
    		copy_to_user((char *)arg,keypad_matrix,sizeof(keypad_matrix));
		else
   			 return -EINVAL;
		break;
	break; 
	default:
		rc = -EINVAL;
		break;
	}
	return(rc);
}

/*****************************************************************************/

irqreturn_t  keypad_isr(int irq, void *dev_id, struct pt_regs *regs)
{

#if KEYPAD_DEBUG
	printk("keypad_isr(irq=%d)\n", irq);
#endif

/**
 * ISR is a trigger for timer state machine
 */
	KEYPAD_MASK_IRQ();
	restart_timer();
	return IRQ_RETVAL(1);	

return IRQ_RETVAL(1);
}

/* Handle read from proc file */
int readProcFile( char *buffer, char **start, off_t offset, int size, int *eof, void *data ){
	char *temp = buffer;
	int row, column;

	/* Print module configuration */
	temp += sprintf( temp, "rows:    %d\n"
			"columns: %d\n\n",KEYPAD_ROWS, KEYPAD_COLUMNS);

for(row=0;row<KEYPAD_ROWS;row++){
			for(column=0;column<KEYPAD_COLUMNS;column++)
				temp+=sprintf(temp,"%c ",keypad_matrix[row*KEYPAD_COLUMNS+column]);
			temp+=sprintf(temp,"\n");
	}
	return temp - buffer;
}



/*****************************************************************************/

/*
 *	Exported file operations structure for driver...
 */

struct file_operations 	keypad_fops = {
	.owner 	= THIS_MODULE,
	.read 		= keypad_read,	
	.poll			= keypad_poll,		
	.ioctl			=	keypad_ioctl,		
	.open		=  keypad_open,	
	.release	=	keypad_release,	
};

/*****************************************************************************/
static int __init start_module(void)
{
dev_t dev = MKDEV(KEYPAD_MAJOR ,FIRSTKEYMINOR);
int result;

keypad_cdev = cdev_alloc();

result = register_chrdev_region(dev, KEYMINORNUM,DRIVER_NAME);

if(result < 0){
	printk(DRIVER_NAME" driver failed to register Major Number"); 
	return result; 
	}

keypad_cdev->ops = &keypad_fops;
keypad_cdev->owner = THIS_MODULE;

	keypad_isopen = 0;
	new_input = 0;

	init_timer(&poll_timer);
	poll_timer.function = (void *)keypad_timer;
	
#ifdef POLLING	
	poll_timer.expires = jiffies + TIMEOUT;
	add_timer(&poll_timer);
#else
	/* register IRQ */{
	int rc;
	KEYPAD_IRQ_INIT();
	rc = request_irq(KEYPAD_IRQ, keypad_isr, SA_INTERRUPT, DRIVER_NAME, NULL);
	if (rc) {
		printk("Keypad irq not registered.  Error: %d\n", rc);
	}
	}
#endif

memset(keypad_matrix,KEYPAD_UNDEFINED_CHAR,sizeof(keypad_matrix));

if(!create_proc_read_entry( "keypad", 0, 0, readProcFile, NULL ) ){
			printk( KERN_ERR "can't create /proc/keypad\n" );
			return -ENOMEM;
		}

cdev_add(keypad_cdev,dev,KEYMINORNUM);

printk(DRIVER_NAME": Copyright (C) 2000-2001, Greg Ungerer "
	"(gerg@snapgear.com)\n"
	"2.6 Update and SoM support by EMAC.Inc->NZG 2005\n"
	);

return 0;
}

/*****************************************************************************/
static void __exit exit_module(void){
	dev_t dev = MKDEV(KEYPAD_MAJOR ,FIRSTKEYMINOR);
	unregister_chrdev_region(dev, KEYMINORNUM);
	cdev_del(keypad_cdev);
}

module_init(start_module);
module_exit(exit_module);
