/*
 *	Watchdog driver for the legacy Coldfire watchdog interface.
 *
 *      (c) Copyright 2006 EMAC.Inc <support@emacinc.com>
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/init.h>
#include <linux/interrupt.h>

#include <asm/bitops.h>
#include <asm/uaccess.h>

#include <asm/irq.h>
#include <asm/coldfire.h>
#include <asm/m528xsim.h>

#include "cfcore_wdt.h"

MODULE_AUTHOR("EMAC.Inc <support@emacinc.com>");
MODULE_DESCRIPTION("Coldfire Core Watchdog");

MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);

static int margin __initdata = 33;		
module_param(margin, int, 0);
MODULE_PARM_DESC(margin, "Watchdog margin in seconds default = 33");

#define M5282_CLOSE_MAGIC	(0x5afc4453)

static unsigned long cfcorewdt_users;
static int expect_close;
static int boot_status;

/*
 *	Process a watchdog reset interrupt. 
 * Soft resets the processor
 */
static irqreturn_t watchdog_reset(int irq, void *dummy, struct pt_regs *fp){
	 printk("watchdog timeout!\n");
	 HARD_RESET_NOW();
	  for (;;) ;  // hang until reboot
	  return IRQ_HANDLED;
}

/*
 *	Allow only one person to hold it open
 */
static int cfcoredog_open(struct inode *inode, struct file *file)
{
	GENDEBUG("cfcoredog_open\n");
	
	nonseekable_open(inode, file);
	if (test_and_set_bit(1,&cfcorewdt_users))
		return -EBUSY;

	/* Activate M5282 Watchdog timer */
	margin = WATCHDOG_SET_TIMEOUT(margin);
	
	GENDEBUG("Coldfire Core Watchdog Timer: timer margin %d sec\n", margin);
	
	//WDT_RESET_ON_TIMEOUT(); //this appears to not work in the 5282, awaiting Freescale confirmation

	//WDT_TA_BEFORE_RESET(); //this also causes strange problems.
	KICK_WATCHDOG();
	WDT_INTERRUPT_CONFIG();
	WATCHDOG_ENABLE();
	
	/*
	w = (unsigned short*)(MCF_MBAR + MCFSIM_WRRR);
	*w = wdt_reset_ref | 0x0001;
	w = (unsigned short*)(MCF_MBAR + MCFSIM_WCR);
	*w = 0;
	*/
	DUMP_WATCH_REGS();
	return 0;
}

/*
 *	Doesn't currently support turning off the Watchdog.
 */
static int cfcoredog_release(struct inode *inode, struct file *file)
{

	printk(KERN_CRIT "WATCHDOG: WDT device closed unexpectedly.  WDT will not stop!\n");
	
	clear_bit(1, &cfcorewdt_users);
	expect_close = 0;
	return 0;
}

static ssize_t cfcoredog_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	
	if (len) {
		GENDEBUG("cfcoredog keepalive\n");
		KICK_WATCHDOG();
		DUMP_WATCH_REGS();
	}
	return len;
}

static struct watchdog_info ident = {
	.options	= WDIOF_CARDRESET | WDIOF_MAGICCLOSE |
			  WDIOF_SETTIMEOUT | WDIOF_KEEPALIVEPING,
	.identity	= "MCF5282 Watchdog",
};

static int cfcoredog_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	int ret = -ENOIOCTLCMD;
	int time;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		ret = copy_to_user((struct watchdog_info *)arg, &ident,
				   sizeof(ident)) ? -EFAULT : 0;
		break;

	case WDIOC_GETSTATUS:
		ret = put_user(0, (int *)arg);
		break;

	case WDIOC_GETBOOTSTATUS:
		ret = put_user(boot_status, (int *)arg);
		break;

	case WDIOC_SETTIMEOUT:
		ret = get_user(time, (int *)arg);
		if (ret)
			break;

		margin = time;
		
		margin = WATCHDOG_SET_TIMEOUT(margin);

		//w = (unsigned short*)(MCF_MBAR + MCFSIM_WRRR);
		//*w = wdt_reset_ref | 0x0001;
		/*fall through*/

	case WDIOC_GETTIMEOUT:
		ret = put_user(margin, (int *)arg);
		break;

	case WDIOC_KEEPALIVE:
		//w = (unsigned short*)(MCF_MBAR + MCFSIM_WCR);
		//*w = 0;
		GENDEBUG("cfcoredog keepalive\n");
		KICK_WATCHDOG();
		DUMP_WATCH_REGS();
		ret = 0;
		break;
	}
	return ret;
}

static struct file_operations cfcoredog_fops =
{
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= cfcoredog_write,
	.ioctl			= cfcoredog_ioctl,
	.open		= cfcoredog_open,
	.release	= cfcoredog_release,
};

static struct miscdevice cfcoredog_miscdev =
{
	.minor		= WATCHDOG_MINOR,
	.name		= "coldfire core watchdog",
	.fops		= &cfcoredog_fops,
};


static int __init cfcoredog_init(void)
{
	int ret;

	/*
	 * Read the reset status, and save it for later.  If
	 * we suspend, RCSR will be cleared, and the watchdog
	 * reset reason will be lost.
	 */
//	boot_status = (RCSR & RCSR_WDR) ? WDIOF_CARDRESET : 0;


	ret = misc_register(&cfcoredog_miscdev);

	request_irq(MCFVEC_CWDT, watchdog_reset, SA_INTERRUPT,
		    "Watchdog Timer", &cfcoredog_miscdev);

	return ret;
}

static void __exit cfcoredog_exit(void){
	misc_deregister(&cfcoredog_miscdev);
}

module_init(cfcoredog_init);
module_exit(cfcoredog_exit);



