/**************************************************************************
*                                                                                         
*  GCFPLD.h - 2.6 low level interface for general IO transactions between the 
* Coldfire EIM and the SoM-100ES CPLD generic bus interface.                                      
*                                                                                         
*  (C) Copyright 2005, EMAC.Inc www.emacinc.com                                           
*      written by              NZG ngustavson@emacinc.com                                 
*                                                                                         
**************************************************************************/
#ifndef _GCFPLD_H_
#define _GCFPLD_H_

#include <asm/types.h>
#include <asm/coldfire.h>              
#include <linux/delay.h>


/******************************************************************************
 * Coldfire EIM bus interface 
 ******************************************************************************/
#define GCFPLDBASE 0x80000000
#define GCFPLDMAJOR 100

#define CSAR3                     *(volatile u16 *)(MCF_MBAR + 0xA4)
#define CSMR3                     *(volatile u32 *)(MCF_MBAR + 0xA8)
	#define VALID						1 
#define CSCR3                     *(volatile u16 *)(MCF_MBAR + 0xAE)
	#define PS16						0xc0
	#define AA							0x100
	#define MAXWAIT					0x3C00


//use 16 bit accesses to use low data lines, doesn't hurt to call this multiple times
#define SETUP_GCF3() {CSAR3=(GCFPLDBASE>>16);CSCR3=PS16|AA;CSMR3=VALID;}
#define SETUP_GCF3SLOW() {CSAR3=(GCFPLDBASE>>16);CSCR3=PS16|AA|MAXWAIT;CSMR3=VALID;}
//GPIO uses low data lines, which will correspond to high data on odd reads
#define ENDREV(index,data) (index%2 ? (data>>8) : data)
#define GCFPLD_WRITE(index,data) *(volatile u8 *)(GCFPLDBASE+index)=(u8)data
#define GCFPLD_READ(index) ENDREV(index,*(volatile u16 *)(GCFPLDBASE+index))


/******************************************************************************
 * CPLD Registers 
 ******************************************************************************/
#define GPIOD0		0
#define GPIOC0		1
#define GPIOD1		2	
#define GPIOC1		3
#define GPIOD2		4
#define GPIOC2		5
#define GPIOD3		6
#define GPIOC3		7
#define KEYPAD		8
#define LCDDREG	 9
#define LCDCREG	10
#define MMCREG		11

 
 /**
  * LCD Control Register bitmaps
  */
#define RW_READ		1
#define RW_WRITE		0	
#define RS_DATA			2
#define RS_INST			0
#define BL_ON				4
#define BL_OFF			0

/**
 * MMC Register bit definitions
 */
#define SDIO_IRQ				0x01
#define CARD_DETECT 	0x02
#define WRITE_PROTECT 0x04


/**
 * Keypad Masks
 */
#define KEYROW				0x03
#define KEYCOLUMN		0x70
#define KEYGATE				0x04
#define ROWCOLMASK	(KEYROW|KEYCOLUMN)


/******************************************************************************
 * GPIO Control
 ******************************************************************************/
/**
 * Low level GPIO read function, calls GCFPLD to access data inside the PLD over
 * the Coldfire bus.
 */
inline u8 GPIO_READ(int index){
SETUP_GCF3();//setup fast chip select access 
return GCFPLD_READ(index);
}

/**
 * Low level GPIO write function, calls GCFPLD to access data inside the PLD over
 * the Coldfire bus.
 */
inline u8 GPIO_WRITE(int index, u8 data){
SETUP_GCF3();//setup fast chip select access 
return GCFPLD_WRITE(index, data);
}

/******************************************************************************
 * LCD Control
 ******************************************************************************/
/**
 * uS delay after asserting RW/data lines before sending data
 */
#define RWDELAY 50

/**
 *  Send a command to the LCD
 */
 inline void LCD_Command(u8 command,int blight){	     
	  SETUP_GCF3SLOW();
	  GCFPLD_WRITE(LCDCREG,(RW_WRITE|RS_INST|blight));	  	  
	  udelay(RWDELAY);
	  GCFPLD_WRITE(LCDDREG,command);
}

/**
 *  Send data to the LCD
 */
inline void LCD_Data(u8 data,int blight){	     
	SETUP_GCF3SLOW();
	GCFPLD_WRITE(LCDCREG,RW_WRITE|RS_DATA|blight);
	udelay(RWDELAY);
	GCFPLD_WRITE(LCDDREG,data);
}

/**
 * Generic LCD bus initialization
 * Initializes the bus interface to the LCD
 * 
 */
inline void LCD_BUSINIT(void){
	//do nothing, chipset parameters set at write time
}

/******************************************************************************
 * MMC Control
 ******************************************************************************/
#define INVALID		0
#define VALID			1
#define CHANGED	2
 
inline int MMC_Detect(int unused){
	SETUP_GCF3();
	return((GCFPLD_READ(MMCREG)&CARD_DETECT ) ? INVALID:VALID);
}

/******************************************************************************
 * Keypad Control
 ******************************************************************************/
inline u8 Keypad_Read(void){
	SETUP_GCF3();	 
	return GCFPLD_READ(KEYPAD);
}

#endif //_GCFPLD_H_





